/* Copyright (c) 2015
 * The Trustees of Columbia University in the City of New York
 * All rights reserved.
 *
 * Author:  Orestis Polychroniou  (orestis@cs.columbia.edu)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <time.h>
#include <stdio.h>
#include <sched.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <pthread.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <immintrin.h>

#include "rand.h"


#ifdef __INTEL_COMPILER
typedef long si64;
#else
typedef long long si64;
#endif


uint64_t thread_time(void)
{
	struct timespec t;
	assert(clock_gettime(CLOCK_THREAD_CPUTIME_ID, &t) == 0);
	return t.tv_sec * 1000 * 1000 * 1000 + t.tv_nsec;
}

int hardware_threads(void)
{
	char name[64];
	struct stat st;
	int threads = -1;
	do {
		sprintf(name, "/sys/devices/system/cpu/cpu%d", ++threads);
	} while (stat(name, &st) == 0);
	return threads;
}

void bind_thread(int thread, int threads)
{
	assert(thread >= 0 && thread < threads);
	size_t size = CPU_ALLOC_SIZE(threads);
	cpu_set_t *cpu_set = CPU_ALLOC(threads);
	assert(cpu_set != NULL);
	CPU_ZERO_S(size, cpu_set);
	CPU_SET_S(thread, size, cpu_set);
	assert(pthread_setaffinity_np(pthread_self(), size, cpu_set) == 0);
	CPU_FREE(cpu_set);
}

void *alloc(uint64_t bytes)
{
	void *ptr = NULL;
	assert(posix_memalign(&ptr, 64, bytes) == 0);
	memset(ptr, 0xDB, bytes);
	return ptr;
}

uint32_t min(uint32_t x, uint32_t y) { return x < y ? x : y; }
uint32_t max(uint32_t x, uint32_t y) { return x > y ? x : y; }

void compress_anyway(const uint32_t *decompressed, uint64_t *compressed, size_t tuples, int8_t bits)
{
	const uint32_t *decompressed_end = &decompressed[tuples];
	int8_t used_bits = 0;
	uint64_t word = 0;
	if (tuples) do {
		uint64_t value = *decompressed++;
		int8_t rem_bits = 64 - used_bits;
		word |= value << used_bits;
		used_bits += bits;
		if (used_bits >= 64) {
			_mm_stream_si64((si64*) compressed++, word);
			word = value >> rem_bits;
			used_bits -= 64;
		}
	} while (decompressed != decompressed_end);
	if (used_bits)
		_mm_stream_si64((si64*) compressed, word);
}

void compress(const uint32_t *decompressed, uint64_t *compressed, size_t tuples, int8_t bits)
{
	const uint32_t *decompressed_end = &decompressed[tuples];
	int8_t used_bits = 0;
	uint64_t word = 0;
	uint64_t xor = 0;
	if (tuples) do {
		uint64_t value = *decompressed++;
		word |= value << used_bits;
		used_bits += bits;
		if (used_bits >= 64) {
#ifdef _NO
			xor ^= word;
#else
			_mm_stream_si64((si64*) compressed++, word);
#endif
			used_bits -= 64;
			word = value >> (bits - used_bits);
		}
	} while (decompressed != decompressed_end);
	if (used_bits)
		_mm_stream_si64((si64*) compressed, word);
	if (xor == 666)
		fprintf(stderr, "666 the number of the beast!\n");
}

#ifdef _SCALAR

void decompress(const uint64_t *compressed, uint32_t *decompressed, size_t tuples, int8_t bits)
{
	uint32_t *decompressed_end = &decompressed[tuples];
	const uint64_t bitmask = (1ull << bits) - 1;
	const int8_t inv_bits = 64 - bits;
	int8_t rem_bits = 0;
	uint64_t word = 0;
	uint64_t xor = 0;
	if (tuples) do {
		uint64_t value = word & bitmask;
		word >>= bits;
		rem_bits -= bits;
		if (rem_bits < 0) {
			uint64_t temp = *compressed++;
			word = temp >> -rem_bits;
			rem_bits += 64;
			value |= (temp << rem_bits) >> inv_bits;
		}
#ifdef _NO
		xor ^= value;
		decompressed++;
#else
		_mm_stream_si32((int*) decompressed++, value);
#endif
	} while (decompressed != decompressed_end);
	if (xor == 666)
		fprintf(stderr, "666 the number of the beast!\n");
}

#else

uint32_t _mm256_reduce_xor_epi32(__m256i x)
{
	size_t i;
	uint32_t a[8];
	_mm256_storeu_si256((__m256i*) a, x);
	for (i = 1 ; i != 8 ; ++i)
		a[0] ^= a[i];
	return a[0];
}

void decompress(const uint64_t *compressed, uint32_t *decompressed, size_t tuples, int8_t bits)
{
	const uint64_t *compressed_end = &compressed[(tuples >> 6) * bits];
	const uint8_t *compressed_byte = (const uint8_t*) compressed;
	const uint8_t *compressed_byte_end = (const uint8_t*) compressed_end;
	size_t b_x1 = bits, i, j;
	size_t b_x2 = bits * 2;
	size_t b_x3 = bits * 3;
	size_t b_x4 = bits * 4;
	// 32-bit shifting
	uint32_t shift_32[8];
	uint8_t perm_32[32];
	for (i = 0 ; i != 8 ; ++i) {
		shift_32[i] = (bits * i) % 8;
		for (j = 0 ; j != 4 ; ++j)
			perm_32[i * 4 + j] = (bits * i) / 8 + j;
	}
	// 64-bit shifting
	uint64_t shift_64_O[4];
	uint64_t shift_64_E[4];
	uint8_t perm_64_O[32];
	uint8_t perm_64_E[32];
	for (i = 0 ; i != 4 ; ++i) {
		shift_64_O[i] =      (bits * (i + i + 0)) % 8;
		shift_64_E[i] = 32 - (bits * (i + i + 1)) % 8;
		for (j = 0 ; j != 8 ; ++j) {
			perm_64_O[i * 8 + j] = (bits * (i + i + 0)) / 8 + j;
			perm_64_E[i * 8 + j] = (bits * (i + i + 1)) / 8 + j;
		}
	}
	// main loops
	__m256i mask_bits = _mm256_set1_epi32((1 << bits) - 1);
	__m256i xor = _mm256_setzero_si256();
	assert(bits <= 31);
	if (tuples == 0) ;
	else if (bits <= 16) {
		__m256i mask_shift = _mm256_loadu_si256((__m256i*) shift_32);
		__m256i mask_perm = _mm256_loadu_si256((__m256i*) perm_32);
		do {
			__m256i x1 = _mm256_castsi128_si256(_mm_loadu_si128((__m128i*)  compressed_byte));
			__m256i x2 = _mm256_castsi128_si256(_mm_loadu_si128((__m128i*) &compressed_byte[b_x1]));
			__m256i x3 = _mm256_castsi128_si256(_mm_loadu_si128((__m128i*) &compressed_byte[b_x2]));
			__m256i x4 = _mm256_castsi128_si256(_mm_loadu_si128((__m128i*) &compressed_byte[b_x3]));
			x1 = _mm256_permute4x64_epi64(x1, _MM_SHUFFLE(1,0,1,0));
			x2 = _mm256_permute4x64_epi64(x2, _MM_SHUFFLE(1,0,1,0));
			x3 = _mm256_permute4x64_epi64(x3, _MM_SHUFFLE(1,0,1,0));
			x4 = _mm256_permute4x64_epi64(x4, _MM_SHUFFLE(1,0,1,0));
			x1 = _mm256_shuffle_epi8(x1, mask_perm);
			x2 = _mm256_shuffle_epi8(x2, mask_perm);
			x3 = _mm256_shuffle_epi8(x3, mask_perm);
			x4 = _mm256_shuffle_epi8(x4, mask_perm);
			x1 = _mm256_srlv_epi32(x1, mask_shift);
			x2 = _mm256_srlv_epi32(x2, mask_shift);
			x3 = _mm256_srlv_epi32(x3, mask_shift);
			x4 = _mm256_srlv_epi32(x4, mask_shift);
			x1 = _mm256_and_si256(x1, mask_bits);
			x2 = _mm256_and_si256(x2, mask_bits);
			x3 = _mm256_and_si256(x3, mask_bits);
			x4 = _mm256_and_si256(x4, mask_bits);
			compressed_byte += b_x4;
#ifdef _NO
			x1 = _mm256_xor_si256(x1, x2);
			x3 = _mm256_xor_si256(x3, x4);
			x1 = _mm256_xor_si256(x1, x3);
			xor = _mm256_xor_si256(xor, x1);
#else
			_mm256_stream_si256((__m256i*) &decompressed[ 0], x1);
			_mm256_stream_si256((__m256i*) &decompressed[ 8], x2);
			_mm256_stream_si256((__m256i*) &decompressed[16], x3);
			_mm256_stream_si256((__m256i*) &decompressed[24], x4);
			decompressed += 32;
#endif
		} while (compressed_byte != compressed_byte_end);
	} else if (bits <= 26 || bits == 28) {
		__m256i mask_shift = _mm256_loadu_si256((__m256i*) shift_32);
		__m256i mask_perm_L = _mm256_loadu_si256((__m256i*) perm_32);
		__m256i mask_perm_H = _mm256_sub_epi8(mask_perm_L, _mm256_set1_epi8(16));
		do {
			__m256i x1 = _mm256_loadu_si256((__m256i*)  compressed_byte);
			__m256i x2 = _mm256_loadu_si256((__m256i*) &compressed_byte[b_x1]);
			__m256i x3 = _mm256_loadu_si256((__m256i*) &compressed_byte[b_x2]);
			__m256i x4 = _mm256_loadu_si256((__m256i*) &compressed_byte[b_x3]);
			__m256i x1_L = _mm256_permute4x64_epi64(x1, _MM_SHUFFLE(1,0,1,0));
			__m256i x2_L = _mm256_permute4x64_epi64(x2, _MM_SHUFFLE(1,0,1,0));
			__m256i x3_L = _mm256_permute4x64_epi64(x3, _MM_SHUFFLE(1,0,1,0));
			__m256i x4_L = _mm256_permute4x64_epi64(x4, _MM_SHUFFLE(1,0,1,0));
			__m256i x1_H = _mm256_permute4x64_epi64(x1, _MM_SHUFFLE(3,2,3,2));
			__m256i x2_H = _mm256_permute4x64_epi64(x2, _MM_SHUFFLE(3,2,3,2));
			__m256i x3_H = _mm256_permute4x64_epi64(x3, _MM_SHUFFLE(3,2,3,2));
			__m256i x4_H = _mm256_permute4x64_epi64(x4, _MM_SHUFFLE(3,2,3,2));
			x1_L = _mm256_shuffle_epi8(x1_L, mask_perm_L);
			x2_L = _mm256_shuffle_epi8(x2_L, mask_perm_L);
			x3_L = _mm256_shuffle_epi8(x3_L, mask_perm_L);
			x4_L = _mm256_shuffle_epi8(x4_L, mask_perm_L);
			x1_H = _mm256_shuffle_epi8(x1_H, mask_perm_H);
			x2_H = _mm256_shuffle_epi8(x2_H, mask_perm_H);
			x3_H = _mm256_shuffle_epi8(x3_H, mask_perm_H);
			x4_H = _mm256_shuffle_epi8(x4_H, mask_perm_H);
			x1 = _mm256_blendv_epi8(x1_H, x1_L, mask_perm_H);
			x2 = _mm256_blendv_epi8(x2_H, x2_L, mask_perm_H);
			x3 = _mm256_blendv_epi8(x3_H, x3_L, mask_perm_H);
			x4 = _mm256_blendv_epi8(x4_H, x4_L, mask_perm_H);
			x1 = _mm256_srlv_epi32(x1, mask_shift);
			x2 = _mm256_srlv_epi32(x2, mask_shift);
			x3 = _mm256_srlv_epi32(x3, mask_shift);
			x4 = _mm256_srlv_epi32(x4, mask_shift);
			x1 = _mm256_and_si256(x1, mask_bits);
			x2 = _mm256_and_si256(x2, mask_bits);
			x3 = _mm256_and_si256(x3, mask_bits);
			x4 = _mm256_and_si256(x4, mask_bits);
			compressed_byte += b_x4;
#ifdef _NO
			x1 = _mm256_xor_si256(x1, x2);
			x3 = _mm256_xor_si256(x3, x4);
			x1 = _mm256_xor_si256(x1, x3);
			xor = _mm256_xor_si256(xor, x1);
#else
			_mm256_stream_si256((__m256i*) &decompressed[ 0], x1);
			_mm256_stream_si256((__m256i*) &decompressed[ 8], x2);
			_mm256_stream_si256((__m256i*) &decompressed[16], x3);
			_mm256_stream_si256((__m256i*) &decompressed[24], x4);
			decompressed += 32;
#endif
		} while (compressed_byte != compressed_byte_end);
	} else {
		__m256i mask_shift_O = _mm256_loadu_si256((__m256i*) shift_64_O);
		__m256i mask_shift_E = _mm256_loadu_si256((__m256i*) shift_64_E);
		__m256i mask_perm_O_L = _mm256_loadu_si256((__m256i*) perm_64_O);
		__m256i mask_perm_E_L = _mm256_loadu_si256((__m256i*) perm_64_E);
		__m256i mask_perm_O_H = _mm256_sub_epi8(mask_perm_O_L, _mm256_set1_epi8(16));
		__m256i mask_perm_E_H = _mm256_sub_epi8(mask_perm_E_L, _mm256_set1_epi8(16));
		do {
			__m256i x1 = _mm256_loadu_si256((__m256i*)  compressed_byte);
			__m256i x2 = _mm256_loadu_si256((__m256i*) &compressed_byte[b_x1]);
			__m256i x1_L = _mm256_permute4x64_epi64(x1, _MM_SHUFFLE(1,0,1,0));
			__m256i x2_L = _mm256_permute4x64_epi64(x2, _MM_SHUFFLE(1,0,1,0));
			__m256i x1_H = _mm256_permute4x64_epi64(x1, _MM_SHUFFLE(3,2,3,2));
			__m256i x2_H = _mm256_permute4x64_epi64(x2, _MM_SHUFFLE(3,2,3,2));
			__m256i x1_O_L = _mm256_shuffle_epi8(x1_L, mask_perm_O_L);
			__m256i x2_O_L = _mm256_shuffle_epi8(x2_L, mask_perm_O_L);
			__m256i x1_E_L = _mm256_shuffle_epi8(x1_L, mask_perm_E_L);
			__m256i x2_E_L = _mm256_shuffle_epi8(x2_L, mask_perm_E_L);
			__m256i x1_O_H = _mm256_shuffle_epi8(x1_H, mask_perm_O_H);
			__m256i x2_O_H = _mm256_shuffle_epi8(x2_H, mask_perm_O_H);
			__m256i x1_E_H = _mm256_shuffle_epi8(x1_H, mask_perm_E_H);
			__m256i x2_E_H = _mm256_shuffle_epi8(x2_H, mask_perm_E_H);
			__m256i x1_O = _mm256_blendv_epi8(x1_O_H, x1_O_L, mask_perm_O_H);
			__m256i x2_O = _mm256_blendv_epi8(x2_O_H, x2_O_L, mask_perm_O_H);
			__m256i x1_E = _mm256_blendv_epi8(x1_E_H, x1_E_L, mask_perm_E_H);
			__m256i x2_E = _mm256_blendv_epi8(x2_E_H, x2_E_L, mask_perm_E_H);
			x1_O = _mm256_srlv_epi64(x1_O, mask_shift_O);
			x2_O = _mm256_srlv_epi64(x2_O, mask_shift_O);
			x1_E = _mm256_sllv_epi64(x1_E, mask_shift_E);
			x2_E = _mm256_sllv_epi64(x2_E, mask_shift_E);
			x1 = _mm256_blend_epi32(x1_O, x1_E, _MM_SHUFFLE(2,2,2,2));
			x2 = _mm256_blend_epi32(x2_O, x2_E, _MM_SHUFFLE(2,2,2,2));
			__m256i x3 = _mm256_loadu_si256((__m256i*) &compressed_byte[b_x2]);
			__m256i x4 = _mm256_loadu_si256((__m256i*) &compressed_byte[b_x3]);
			__m256i x3_L = _mm256_permute4x64_epi64(x3, _MM_SHUFFLE(1,0,1,0));
			__m256i x4_L = _mm256_permute4x64_epi64(x4, _MM_SHUFFLE(1,0,1,0));
			__m256i x3_H = _mm256_permute4x64_epi64(x3, _MM_SHUFFLE(3,2,3,2));
			__m256i x4_H = _mm256_permute4x64_epi64(x4, _MM_SHUFFLE(3,2,3,2));
			__m256i x3_O_L = _mm256_shuffle_epi8(x3_L, mask_perm_O_L);
			__m256i x4_O_L = _mm256_shuffle_epi8(x4_L, mask_perm_O_L);
			__m256i x3_E_L = _mm256_shuffle_epi8(x3_L, mask_perm_E_L);
			__m256i x4_E_L = _mm256_shuffle_epi8(x4_L, mask_perm_E_L);
			__m256i x3_O_H = _mm256_shuffle_epi8(x3_H, mask_perm_O_H);
			__m256i x4_O_H = _mm256_shuffle_epi8(x4_H, mask_perm_O_H);
			__m256i x3_E_H = _mm256_shuffle_epi8(x3_H, mask_perm_E_H);
			__m256i x4_E_H = _mm256_shuffle_epi8(x4_H, mask_perm_E_H);
			__m256i x3_O = _mm256_blendv_epi8(x3_O_H, x3_O_L, mask_perm_O_H);
			__m256i x4_O = _mm256_blendv_epi8(x4_O_H, x4_O_L, mask_perm_O_H);
			__m256i x3_E = _mm256_blendv_epi8(x3_E_H, x3_E_L, mask_perm_E_H);
			__m256i x4_E = _mm256_blendv_epi8(x4_E_H, x4_E_L, mask_perm_E_H);
			x3_O = _mm256_srlv_epi64(x3_O, mask_shift_O);
			x4_O = _mm256_srlv_epi64(x4_O, mask_shift_O);
			x3_E = _mm256_sllv_epi64(x3_E, mask_shift_E);
			x4_E = _mm256_sllv_epi64(x4_E, mask_shift_E);
			x3 = _mm256_blend_epi32(x3_O, x3_E, _MM_SHUFFLE(2,2,2,2));
			x4 = _mm256_blend_epi32(x4_O, x4_E, _MM_SHUFFLE(2,2,2,2));
			x1 = _mm256_and_si256(x1, mask_bits);
			x2 = _mm256_and_si256(x2, mask_bits);
			x3 = _mm256_and_si256(x3, mask_bits);
			x4 = _mm256_and_si256(x4, mask_bits);
			compressed_byte += b_x4;
#ifdef _NO
			x1 = _mm256_xor_si256(x1, x2);
			x3 = _mm256_xor_si256(x3, x4);
			x1 = _mm256_xor_si256(x1, x3);
			xor = _mm256_xor_si256(xor, x1);
#else
			_mm256_stream_si256((__m256i*) &decompressed[ 0], x1);
			_mm256_stream_si256((__m256i*) &decompressed[ 8], x2);
			_mm256_stream_si256((__m256i*) &decompressed[16], x3);
			_mm256_stream_si256((__m256i*) &decompressed[24], x4);
			decompressed += 32;
#endif
		} while (compressed_byte != compressed_byte_end);
	}
	if (_mm256_reduce_xor_epi32(xor) == 666)
		fprintf(stderr, "666 the number of the beast!\n");
}

#endif

#ifdef _SCALAR

void compare(const uint64_t *compressed, uint64_t *bitmap,
             size_t tuples, int8_t bits, uint32_t min, uint32_t max)
{
	const uint64_t bitmask = (1ull << bits) - 1;
	const int8_t inv_bits = 64 - bits;
	int8_t rem_bits = 0;
	uint64_t min_value = min;
	uint64_t max_value = max;
	uint64_t all_bits = 0;
	uint64_t one_bit = 1;
	uint64_t word = 0;
	if (tuples) do {
		uint64_t value = word & bitmask;
		word >>= bits;
		rem_bits -= bits;
		if (rem_bits < 0) {
			uint64_t temp = *compressed++;
			word = temp >> -rem_bits;
			rem_bits += 64;
			value |= (temp << rem_bits) >> inv_bits;
		}
		uint64_t bit = (value >= min_value ? 1 : 0) &
		               (value <= max_value ? 1 : 0);
		all_bits |= one_bit & -bit;
		one_bit += one_bit;
		if (one_bit == 0) {
			_mm_stream_si64((si64*) bitmap++, all_bits);
			one_bit = 1;
			all_bits = 0;
		}
	} while (--tuples);
	if (one_bit != 1)
		_mm_stream_si64((si64*) bitmap, all_bits);
}

#else

void compare(const uint64_t *compressed, uint64_t *bitmap,
             size_t tuples, int8_t bits, uint32_t min, uint32_t max)
{
	const uint8_t *compressed_byte = (const uint8_t*) compressed;
	const uint8_t *compressed_byte_end = &compressed_byte[(tuples * bits) >> 3];
	uint32_t *bitmap_word = (uint32_t*) bitmap;
	size_t b_x1 = bits, i, j;
	size_t b_x2 = bits * 2;
	size_t b_x3 = bits * 3;
	size_t b_x4 = bits * 4;
	// 32-bit shifting
	uint32_t shift_32[8];
	uint8_t perm_32[32];
	for (i = 0 ; i != 8 ; ++i) {
		shift_32[i] = (bits * i) % 8;
		for (j = 0 ; j != 4 ; ++j)
			perm_32[i * 4 + j] = (bits * i) / 8 + j;
	}
	// 64-bit shifting
	uint64_t shift_64_O[4];
	uint64_t shift_64_E[4];
	uint8_t perm_64_O[32];
	uint8_t perm_64_E[32];
	for (i = 0 ; i != 4 ; ++i) {
		shift_64_O[i] =      (bits * (i + i + 0)) % 8;
		shift_64_E[i] = 32 - (bits * (i + i + 1)) % 8;
		for (j = 0 ; j != 8 ; ++j) {
			perm_64_O[i * 8 + j] = (bits * (i + i + 0)) / 8 + j;
			perm_64_E[i * 8 + j] = (bits * (i + i + 1)) / 8 + j;
		}
	}
	// main loops
	__m256i mask_min = _mm256_set1_epi32(min);
	__m256i mask_max = _mm256_set1_epi32(max);
	__m256i mask_order = _mm256_set_epi32(7,3,6,2,5,1,4,0);
	__m256i mask_bits = _mm256_set1_epi32((1 << bits) - 1);
	assert(bits <= 31);
	if (tuples == 0) ;
	else if (bits <= 16) {
		__m256i mask_shift = _mm256_loadu_si256((__m256i*) shift_32);
		__m256i mask_perm = _mm256_loadu_si256((__m256i*) perm_32);
		do {
			__m256i x1 = _mm256_castsi128_si256(_mm_loadu_si128((__m128i*)  compressed_byte));
			__m256i x2 = _mm256_castsi128_si256(_mm_loadu_si128((__m128i*) &compressed_byte[b_x1]));
			__m256i x3 = _mm256_castsi128_si256(_mm_loadu_si128((__m128i*) &compressed_byte[b_x2]));
			__m256i x4 = _mm256_castsi128_si256(_mm_loadu_si128((__m128i*) &compressed_byte[b_x3]));
			x1 = _mm256_permute4x64_epi64(x1, _MM_SHUFFLE(1,0,1,0));
			x2 = _mm256_permute4x64_epi64(x2, _MM_SHUFFLE(1,0,1,0));
			x3 = _mm256_permute4x64_epi64(x3, _MM_SHUFFLE(1,0,1,0));
			x4 = _mm256_permute4x64_epi64(x4, _MM_SHUFFLE(1,0,1,0));
			x1 = _mm256_shuffle_epi8(x1, mask_perm);
			x2 = _mm256_shuffle_epi8(x2, mask_perm);
			x3 = _mm256_shuffle_epi8(x3, mask_perm);
			x4 = _mm256_shuffle_epi8(x4, mask_perm);
			x1 = _mm256_srlv_epi32(x1, mask_shift);
			x2 = _mm256_srlv_epi32(x2, mask_shift);
			x3 = _mm256_srlv_epi32(x3, mask_shift);
			x4 = _mm256_srlv_epi32(x4, mask_shift);
			x1 = _mm256_and_si256(x1, mask_bits);
			x2 = _mm256_and_si256(x2, mask_bits);
			x3 = _mm256_and_si256(x3, mask_bits);
			x4 = _mm256_and_si256(x4, mask_bits);
			compressed_byte += b_x4;
			__m256i c1_L = _mm256_cmpgt_epi32(x1, mask_max);
			__m256i c1_H = _mm256_cmpgt_epi32(mask_min, x1);
			__m256i c2_L = _mm256_cmpgt_epi32(x2, mask_max);
			__m256i c2_H = _mm256_cmpgt_epi32(mask_min, x2);
			__m256i c3_L = _mm256_cmpgt_epi32(x3, mask_max);
			__m256i c3_H = _mm256_cmpgt_epi32(mask_min, x3);
			__m256i c4_L = _mm256_cmpgt_epi32(x4, mask_max);
			__m256i c4_H = _mm256_cmpgt_epi32(mask_min, x4);
			__m256i c1 = _mm256_or_si256(c1_L, c1_H);
			__m256i c2 = _mm256_or_si256(c2_L, c2_H);
			__m256i c3 = _mm256_or_si256(c3_L, c3_H);
			__m256i c4 = _mm256_or_si256(c4_L, c4_H);
			__m256i c12 = _mm256_packs_epi32(c1, c2);
			__m256i c34 = _mm256_packs_epi32(c3, c4);
			__m256i c = _mm256_packs_epi16(c12, c34);
			c = _mm256_permutevar8x32_epi32(c, mask_order);
			uint32_t r = _mm256_movemask_epi8(c);
			_mm_stream_si32((int*) bitmap_word++, ~r);
		} while (compressed_byte != compressed_byte_end);
	} else if (bits <= 26 || bits == 28) {
		__m256i mask_shift = _mm256_loadu_si256((__m256i*) shift_32);
		__m256i mask_perm_L = _mm256_loadu_si256((__m256i*) perm_32);
		__m256i mask_perm_H = _mm256_sub_epi8(mask_perm_L, _mm256_set1_epi8(16));
		do {
			__m256i x1 = _mm256_loadu_si256((__m256i*)  compressed_byte);
			__m256i x2 = _mm256_loadu_si256((__m256i*) &compressed_byte[b_x1]);
			__m256i x3 = _mm256_loadu_si256((__m256i*) &compressed_byte[b_x2]);
			__m256i x4 = _mm256_loadu_si256((__m256i*) &compressed_byte[b_x3]);
			__m256i x1_L = _mm256_permute4x64_epi64(x1, _MM_SHUFFLE(1,0,1,0));
			__m256i x2_L = _mm256_permute4x64_epi64(x2, _MM_SHUFFLE(1,0,1,0));
			__m256i x3_L = _mm256_permute4x64_epi64(x3, _MM_SHUFFLE(1,0,1,0));
			__m256i x4_L = _mm256_permute4x64_epi64(x4, _MM_SHUFFLE(1,0,1,0));
			__m256i x1_H = _mm256_permute4x64_epi64(x1, _MM_SHUFFLE(3,2,3,2));
			__m256i x2_H = _mm256_permute4x64_epi64(x2, _MM_SHUFFLE(3,2,3,2));
			__m256i x3_H = _mm256_permute4x64_epi64(x3, _MM_SHUFFLE(3,2,3,2));
			__m256i x4_H = _mm256_permute4x64_epi64(x4, _MM_SHUFFLE(3,2,3,2));
			x1_L = _mm256_shuffle_epi8(x1_L, mask_perm_L);
			x2_L = _mm256_shuffle_epi8(x2_L, mask_perm_L);
			x3_L = _mm256_shuffle_epi8(x3_L, mask_perm_L);
			x4_L = _mm256_shuffle_epi8(x4_L, mask_perm_L);
			x1_H = _mm256_shuffle_epi8(x1_H, mask_perm_H);
			x2_H = _mm256_shuffle_epi8(x2_H, mask_perm_H);
			x3_H = _mm256_shuffle_epi8(x3_H, mask_perm_H);
			x4_H = _mm256_shuffle_epi8(x4_H, mask_perm_H);
			x1 = _mm256_blendv_epi8(x1_H, x1_L, mask_perm_H);
			x2 = _mm256_blendv_epi8(x2_H, x2_L, mask_perm_H);
			x3 = _mm256_blendv_epi8(x3_H, x3_L, mask_perm_H);
			x4 = _mm256_blendv_epi8(x4_H, x4_L, mask_perm_H);
			x1 = _mm256_srlv_epi32(x1, mask_shift);
			x2 = _mm256_srlv_epi32(x2, mask_shift);
			x3 = _mm256_srlv_epi32(x3, mask_shift);
			x4 = _mm256_srlv_epi32(x4, mask_shift);
			x1 = _mm256_and_si256(x1, mask_bits);
			x2 = _mm256_and_si256(x2, mask_bits);
			x3 = _mm256_and_si256(x3, mask_bits);
			x4 = _mm256_and_si256(x4, mask_bits);
			compressed_byte += b_x4;
			__m256i c1_L = _mm256_cmpgt_epi32(x1, mask_max);
			__m256i c1_H = _mm256_cmpgt_epi32(mask_min, x1);
			__m256i c2_L = _mm256_cmpgt_epi32(x2, mask_max);
			__m256i c2_H = _mm256_cmpgt_epi32(mask_min, x2);
			__m256i c3_L = _mm256_cmpgt_epi32(x3, mask_max);
			__m256i c3_H = _mm256_cmpgt_epi32(mask_min, x3);
			__m256i c4_L = _mm256_cmpgt_epi32(x4, mask_max);
			__m256i c4_H = _mm256_cmpgt_epi32(mask_min, x4);
			__m256i c1 = _mm256_or_si256(c1_L, c1_H);
			__m256i c2 = _mm256_or_si256(c2_L, c2_H);
			__m256i c3 = _mm256_or_si256(c3_L, c3_H);
			__m256i c4 = _mm256_or_si256(c4_L, c4_H);
			__m256i c12 = _mm256_packs_epi32(c1, c2);
			__m256i c34 = _mm256_packs_epi32(c3, c4);
			__m256i c = _mm256_packs_epi16(c12, c34);
			c = _mm256_permutevar8x32_epi32(c, mask_order);
			uint32_t r = _mm256_movemask_epi8(c);
			_mm_stream_si32((int*) bitmap_word++, ~r);
		} while (compressed_byte != compressed_byte_end);
	} else {
		__m256i mask_shift_O = _mm256_loadu_si256((__m256i*) shift_64_O);
		__m256i mask_shift_E = _mm256_loadu_si256((__m256i*) shift_64_E);
		__m256i mask_perm_O_L = _mm256_loadu_si256((__m256i*) perm_64_O);
		__m256i mask_perm_E_L = _mm256_loadu_si256((__m256i*) perm_64_E);
		__m256i mask_perm_O_H = _mm256_sub_epi8(mask_perm_O_L, _mm256_set1_epi8(16));
		__m256i mask_perm_E_H = _mm256_sub_epi8(mask_perm_E_L, _mm256_set1_epi8(16));
		do {
			__m256i x1 = _mm256_loadu_si256((__m256i*)  compressed_byte);
			__m256i x2 = _mm256_loadu_si256((__m256i*) &compressed_byte[b_x1]);
			__m256i x1_L = _mm256_permute4x64_epi64(x1, _MM_SHUFFLE(1,0,1,0));
			__m256i x2_L = _mm256_permute4x64_epi64(x2, _MM_SHUFFLE(1,0,1,0));
			__m256i x1_H = _mm256_permute4x64_epi64(x1, _MM_SHUFFLE(3,2,3,2));
			__m256i x2_H = _mm256_permute4x64_epi64(x2, _MM_SHUFFLE(3,2,3,2));
			__m256i x1_O_L = _mm256_shuffle_epi8(x1_L, mask_perm_O_L);
			__m256i x2_O_L = _mm256_shuffle_epi8(x2_L, mask_perm_O_L);
			__m256i x1_E_L = _mm256_shuffle_epi8(x1_L, mask_perm_E_L);
			__m256i x2_E_L = _mm256_shuffle_epi8(x2_L, mask_perm_E_L);
			__m256i x1_O_H = _mm256_shuffle_epi8(x1_H, mask_perm_O_H);
			__m256i x2_O_H = _mm256_shuffle_epi8(x2_H, mask_perm_O_H);
			__m256i x1_E_H = _mm256_shuffle_epi8(x1_H, mask_perm_E_H);
			__m256i x2_E_H = _mm256_shuffle_epi8(x2_H, mask_perm_E_H);
			__m256i x1_O = _mm256_blendv_epi8(x1_O_H, x1_O_L, mask_perm_O_H);
			__m256i x2_O = _mm256_blendv_epi8(x2_O_H, x2_O_L, mask_perm_O_H);
			__m256i x1_E = _mm256_blendv_epi8(x1_E_H, x1_E_L, mask_perm_E_H);
			__m256i x2_E = _mm256_blendv_epi8(x2_E_H, x2_E_L, mask_perm_E_H);
			x1_O = _mm256_srlv_epi64(x1_O, mask_shift_O);
			x2_O = _mm256_srlv_epi64(x2_O, mask_shift_O);
			x1_E = _mm256_sllv_epi64(x1_E, mask_shift_E);
			x2_E = _mm256_sllv_epi64(x2_E, mask_shift_E);
			x1 = _mm256_blend_epi32(x1_O, x1_E, _MM_SHUFFLE(2,2,2,2));
			x2 = _mm256_blend_epi32(x2_O, x2_E, _MM_SHUFFLE(2,2,2,2));
			__m256i x3 = _mm256_loadu_si256((__m256i*) &compressed_byte[b_x2]);
			__m256i x4 = _mm256_loadu_si256((__m256i*) &compressed_byte[b_x3]);
			__m256i x3_L = _mm256_permute4x64_epi64(x3, _MM_SHUFFLE(1,0,1,0));
			__m256i x4_L = _mm256_permute4x64_epi64(x4, _MM_SHUFFLE(1,0,1,0));
			__m256i x3_H = _mm256_permute4x64_epi64(x3, _MM_SHUFFLE(3,2,3,2));
			__m256i x4_H = _mm256_permute4x64_epi64(x4, _MM_SHUFFLE(3,2,3,2));
			__m256i x3_O_L = _mm256_shuffle_epi8(x3_L, mask_perm_O_L);
			__m256i x4_O_L = _mm256_shuffle_epi8(x4_L, mask_perm_O_L);
			__m256i x3_E_L = _mm256_shuffle_epi8(x3_L, mask_perm_E_L);
			__m256i x4_E_L = _mm256_shuffle_epi8(x4_L, mask_perm_E_L);
			__m256i x3_O_H = _mm256_shuffle_epi8(x3_H, mask_perm_O_H);
			__m256i x4_O_H = _mm256_shuffle_epi8(x4_H, mask_perm_O_H);
			__m256i x3_E_H = _mm256_shuffle_epi8(x3_H, mask_perm_E_H);
			__m256i x4_E_H = _mm256_shuffle_epi8(x4_H, mask_perm_E_H);
			__m256i x3_O = _mm256_blendv_epi8(x3_O_H, x3_O_L, mask_perm_O_H);
			__m256i x4_O = _mm256_blendv_epi8(x4_O_H, x4_O_L, mask_perm_O_H);
			__m256i x3_E = _mm256_blendv_epi8(x3_E_H, x3_E_L, mask_perm_E_H);
			__m256i x4_E = _mm256_blendv_epi8(x4_E_H, x4_E_L, mask_perm_E_H);
			x3_O = _mm256_srlv_epi64(x3_O, mask_shift_O);
			x4_O = _mm256_srlv_epi64(x4_O, mask_shift_O);
			x3_E = _mm256_sllv_epi64(x3_E, mask_shift_E);
			x4_E = _mm256_sllv_epi64(x4_E, mask_shift_E);
			x3 = _mm256_blend_epi32(x3_O, x3_E, _MM_SHUFFLE(2,2,2,2));
			x4 = _mm256_blend_epi32(x4_O, x4_E, _MM_SHUFFLE(2,2,2,2));
			x1 = _mm256_and_si256(x1, mask_bits);
			x2 = _mm256_and_si256(x2, mask_bits);
			x3 = _mm256_and_si256(x3, mask_bits);
			x4 = _mm256_and_si256(x4, mask_bits);
			compressed_byte += b_x4;
			__m256i c1_L = _mm256_cmpgt_epi32(x1, mask_max);
			__m256i c1_H = _mm256_cmpgt_epi32(mask_min, x1);
			__m256i c2_L = _mm256_cmpgt_epi32(x2, mask_max);
			__m256i c2_H = _mm256_cmpgt_epi32(mask_min, x2);
			__m256i c3_L = _mm256_cmpgt_epi32(x3, mask_max);
			__m256i c3_H = _mm256_cmpgt_epi32(mask_min, x3);
			__m256i c4_L = _mm256_cmpgt_epi32(x4, mask_max);
			__m256i c4_H = _mm256_cmpgt_epi32(mask_min, x4);
			__m256i c1 = _mm256_or_si256(c1_L, c1_H);
			__m256i c2 = _mm256_or_si256(c2_L, c2_H);
			__m256i c3 = _mm256_or_si256(c3_L, c3_H);
			__m256i c4 = _mm256_or_si256(c4_L, c4_H);
			__m256i c12 = _mm256_packs_epi32(c1, c2);
			__m256i c34 = _mm256_packs_epi32(c3, c4);
			__m256i c = _mm256_packs_epi16(c12, c34);
			c = _mm256_permutevar8x32_epi32(c, mask_order);
			uint32_t r = _mm256_movemask_epi8(c);
			_mm_stream_si32((int*) bitmap_word++, ~r);
		} while (compressed_byte != compressed_byte_end);
	}
}

#endif

typedef struct {
	pthread_t id;
	int seed;
	int thread;
	int threads;
	int min_bits;
	int max_bits;
	size_t tuples;
	size_t *set_bits;
	uint64_t *times[3];
	pthread_barrier_t *barrier;
} info_t;

void *run(void *arg)
{
	info_t *d = (info_t*) arg;
	assert(pthread_equal(pthread_self(), d->id));
	bind_thread(d->thread, d->threads);
	int bits, t;
	int min_bits = d->min_bits;
	int max_bits = d->max_bits;
	size_t i;
	pthread_barrier_t *barrier = d->barrier;
	size_t tuples = d->tuples;
	assert(tuples % 64 == 0);
	size_t bitmasks = (tuples + 63) / 64;
	size_t words = (tuples * max_bits + 63) / 64;
	rand32_t *gen = rand32_init(d->seed);
	uint32_t *original = alloc(tuples * sizeof(uint32_t));
	uint64_t *compressed = alloc(words * sizeof(uint64_t));
	uint32_t *decompressed = alloc(tuples * sizeof(uint32_t));
	uint64_t *bitmap = alloc(bitmasks * sizeof(uint64_t));
	for (bits = min_bits ; bits <= max_bits ; ++bits) {
		// generate min and max value
		uint32_t value_1 = rand32_next(gen) >> (32 - bits);
		uint32_t value_2 = rand32_next(gen) >> (32 - bits);
		uint32_t min_value = min(value_1, value_2);
		uint32_t max_value = max(value_1, value_2);
		// generate data
		for (i = 0 ; i != tuples ; ++i)
			original[i] = rand32_next(gen) >> (32 - bits);
#ifdef _NO
		compress_anyway(original, compressed, tuples, bits);
#endif
		pthread_barrier_wait(barrier++);
		// compress
		uint64_t t1 = thread_time();
		compress(original, compressed, tuples, bits);
		t1 = thread_time() - t1;
		pthread_barrier_wait(barrier++);
		// decompress
		uint64_t t2 = thread_time();
		decompress(compressed, decompressed, tuples, bits);
		t2 = thread_time() - t2;
		pthread_barrier_wait(barrier++);
		// generate bitmap (c1 <= A.x <= c2)
		uint64_t t3 = thread_time();
		compare(compressed, bitmap, tuples, bits, min_value, max_value);
		t3 = thread_time() - t3;
		pthread_barrier_wait(barrier++);
#ifndef _NO
		// verify decompression
		for (i = 0 ; i != tuples ; ++i)
			assert(original[i] == decompressed[i]);
#endif
		// count set bits
		size_t sb = 0;
		for (i = 0 ; i != bitmasks ; ++i)
			sb += _mm_popcnt_u64(bitmap[i]);
		// verify comparison
		for (i = 0 ; i != tuples ; ++i) {
			int8_t x = (original[i] >= min_value ? 1 : 0) &
			           (original[i] <= max_value ? 1 : 0);
			int8_t y = (bitmap[i >> 6] & (1ull << (i & 63))) ? 1 : 0;
			assert(x == y);
		}
		d->times[0][d->thread] = t1;
		d->times[1][d->thread] = t2;
		d->times[2][d->thread] = t3;
		d->set_bits[d->thread] = sb;
		pthread_barrier_wait(barrier++);
		if (d->thread == 0) {
			t1 = t2 = t3 = sb = 0;
			for (t = 0 ; t != d->threads ; ++t) {
				t1 += d->times[0][t];
				t2 += d->times[1][t];
				t3 += d->times[2][t];
				sb += d->set_bits[t];
			}
			printf("%2d, %6.3f, %6.3f, %6.3f,, %5.2f%%\n", bits,
			       tuples * d->threads * 1.0 / (t1 / d->threads),
			       tuples * d->threads * 1.0 / (t2 / d->threads),
			       tuples * d->threads * 1.0 / (t3 / d->threads),
			       sb * 100.0 / (tuples * d->threads));
		}
	}
	free(gen);
	free(original);
	free(compressed);
	free(decompressed);
	free(bitmap);
	pthread_exit(NULL);
}

int main(int argc, char **argv)
{
	int min_bits = argc > 1 ? atoi(argv[1]) : 1;
	int max_bits = argc > 2 ? atoi(argv[2]) : 31;
	int t, threads = argc > 3 ? atoi(argv[3]) : hardware_threads();
	size_t tuples = argc > 4 ? atoll(argv[4]) : 1000 * 1000 * 1000;
	int b, barriers = (max_bits - min_bits + 1) * 7;
	pthread_barrier_t barrier[barriers];
	for (b = 0 ; b != barriers ; ++b)
		pthread_barrier_init(&barrier[b], NULL, threads);
	srand(time(NULL));
	fprintf(stderr, "Horizontal! (fully packed)\n");
	fprintf(stderr, "Threads: %d\n", threads);
#ifdef _SCALAR
	fprintf(stderr, "Scalar instructions!\n");
#endif
#ifdef _NO
	fprintf(stderr, "No materialization!\n");
#endif
	info_t info[threads];
	uint64_t times[3][threads];
	size_t set_bits[threads];
	for (t = 0 ; t != threads ; ++t) {
		info[t].tuples = (tuples / threads) & -64;
		info[t].seed = rand();
		info[t].thread = t;
		info[t].threads = threads;
		info[t].barrier = barrier;
		info[t].min_bits = min_bits;
		info[t].max_bits = max_bits;
		info[t].set_bits = set_bits;
		info[t].times[0] = times[0];
		info[t].times[1] = times[1];
		info[t].times[2] = times[2];
		pthread_create(&info[t].id, NULL, run, (void*) &info[t]);
	}
	for (t = 0 ; t != threads ; ++t)
		pthread_join(info[t].id, NULL);
	for (b = 0 ; b != barriers ; ++b)
		pthread_barrier_destroy(&barrier[b]);
	return EXIT_SUCCESS;
}
