/* Copyright (c) 2015
 * The Trustees of Columbia University in the City of New York
 * All rights reserved.
 *
 * Author:  Orestis Polychroniou  (orestis@cs.columbia.edu)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <time.h>
#include <stdio.h>
#include <sched.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <pthread.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <immintrin.h>

#include "rand.h"


#ifdef __INTEL_COMPILER
typedef long si64;
#else
typedef long long si64;
#endif


uint64_t thread_time(void)
{
	struct timespec t;
	assert(clock_gettime(CLOCK_THREAD_CPUTIME_ID, &t) == 0);
	return t.tv_sec * 1000 * 1000 * 1000 + t.tv_nsec;
}

int hardware_threads(void)
{
	char name[64];
	struct stat st;
	int threads = -1;
	do {
		sprintf(name, "/sys/devices/system/cpu/cpu%d", ++threads);
	} while (stat(name, &st) == 0);
	return threads;
}

void bind_thread(int thread, int threads)
{
	assert(thread >= 0 && thread < threads);
	size_t size = CPU_ALLOC_SIZE(threads);
	cpu_set_t *cpu_set = CPU_ALLOC(threads);
	assert(cpu_set != NULL);
	CPU_ZERO_S(size, cpu_set);
	CPU_SET_S(thread, size, cpu_set);
	assert(pthread_setaffinity_np(pthread_self(), size, cpu_set) == 0);
	CPU_FREE(cpu_set);
}

void *alloc(uint64_t bytes)
{
	void *ptr = NULL;
	assert(posix_memalign(&ptr, 64, bytes) == 0);
	memset(ptr, 0xDB, bytes);
	return ptr;
}

uint64_t repeat(int8_t bits, uint64_t value)
{
	uint64_t result = 0;
	uint64_t limit = 1ull << bits;
	size_t repeats = 64 / bits;
	assert(value < limit);
	while (repeats--) {
		result |= value;
		value <<= bits;
	}
	return result;
}

uint32_t min(uint32_t x, uint32_t y) { return x < y ? x : y; }
uint32_t max(uint32_t x, uint32_t y) { return x > y ? x : y; }

void compress_anyway(const uint32_t *decompressed, uint64_t *compressed, size_t tuples, int8_t bits)
{
	const uint32_t *decompressed_end = &decompressed[tuples];
	size_t group = 64 / (bits + 1);
	assert(group > 1 && tuples % group == 0);
	if (tuples) do {
		const uint32_t *decompressed_group_end = &decompressed[group];
		uint64_t word = *decompressed++;
		int8_t shift = bits + 1;
		do {
			uint64_t in = *decompressed++;
			word |= in << shift;
			shift += bits + 1;
		} while (decompressed != decompressed_group_end);
		_mm_stream_si64((si64*) compressed++, word);
	} while (decompressed != decompressed_end);
}

void compress(const uint32_t *decompressed, uint64_t *compressed, size_t tuples, int8_t bits)
{
	const uint32_t *decompressed_end = &decompressed[tuples];
	size_t group = 64 / ++bits;
	assert(group > 1 && tuples % group == 0);
	uint64_t xor = 0;
	if (tuples) do {
		const uint32_t *decompressed_group_end = &decompressed[group];
		uint64_t word = *decompressed++;
		int8_t shift = bits;
		do {
			uint64_t in = *decompressed++;
			word |= in << shift;
			shift += bits;
		} while (decompressed != decompressed_group_end);
#ifdef _NO
		xor ^= word;
#else
		_mm_stream_si64((si64*) compressed++, word);
#endif
	} while (decompressed != decompressed_end);
	if (xor == 666)
		fprintf(stderr, "666 the number of the beast!\n");
}

void decompress(const uint64_t *compressed, uint32_t *decompressed, size_t tuples, int8_t bits)
{
	const uint32_t *decompressed_end = &decompressed[tuples];
	uint64_t mask = (1ull << bits) - 1;
	size_t group = 64 / ++bits;
	assert(group > 1 && tuples % group == 0);
	uint64_t xor = 0;
	if (tuples) do {
		const uint32_t *decompressed_group_end = &decompressed[group];
		uint64_t word = *compressed++;
		do {
			uint64_t value = word & mask;
			word >>= bits;
#ifdef _NO
			xor ^= value;
#else
			_mm_stream_si32((int*) decompressed, value);
#endif
		} while (++decompressed != decompressed_group_end);
	} while (decompressed != decompressed_end);
	if (xor == 666)
		fprintf(stderr, "666 the number of the beast!\n");
}

void compare(const uint64_t *compressed, uint64_t *bitmap,
             size_t tuples, int8_t bits, uint32_t min, uint32_t max)
{
	// compute groups
	size_t group = 64 / (bits + 1);
	assert(group > 1 && tuples % group == 0);
	const uint64_t *compressed_end = &compressed[tuples / group];
	assert(min < (1 << bits) && max < (1 << bits));
	// generate masks
	uint64_t mask_1_0b = repeat(bits + 1, 1ull << bits);
	uint64_t mask_0_1b = mask_1_0b - (mask_1_0b >> bits);
	uint64_t mask_min  = repeat(bits + 1, min);
	uint64_t mask_max  = repeat(bits + 1, max);
	// bitmap info
	uint64_t bitmask = 0;
	int8_t bitmask_bits_used = 0;
	int8_t bitmask_bits = group;
	if (tuples) do {
		uint64_t word = *compressed++;
		// y + (x ^ 0|1^b), iff x < y
		uint64_t gt_min = mask_min + (word ^ mask_0_1b);
		uint64_t lt_max = word + (mask_max ^ mask_0_1b);
		uint64_t r = _pext_u64(gt_min | lt_max, mask_1_0b);
		// append to bitmap
		int8_t bitmask_bits_rem = 64 - bitmask_bits_used;
		bitmask |= r << bitmask_bits_used;
		bitmask_bits_used += bitmask_bits;
		if (bitmask_bits_rem <= bitmask_bits) {
			_mm_stream_si64((si64*) bitmap++, ~bitmask);
			bitmask = r >> bitmask_bits_rem;
			bitmask_bits_used -= 64;
		}
	} while (compressed != compressed_end);
	if (bitmask_bits_used)
		_mm_stream_si64((si64*) bitmap, ~bitmask);
}

typedef struct {
	pthread_t id;
	int seed;
	int thread;
	int threads;
	int min_bits;
	int max_bits;
	size_t tuples;
	size_t *set_bits;
	uint64_t *times[3];
	pthread_barrier_t *barrier;
} info_t;

void *run(void *arg)
{
	info_t *d = (info_t*) arg;
	assert(pthread_equal(pthread_self(), d->id));
	bind_thread(d->thread, d->threads);
	int bits, t;
	int min_bits = d->min_bits;
	int max_bits = d->max_bits;
	size_t i;
	pthread_barrier_t *barrier = d->barrier;
	size_t tuples = d->tuples;
	assert(tuples % 64 == 0);
	size_t bitmasks = (tuples + 63) / 64;
	rand32_t *gen = rand32_init(d->seed);
	uint32_t *original = alloc(tuples * sizeof(uint32_t));
	uint64_t *compressed = alloc(tuples * sizeof(uint64_t));
	uint32_t *decompressed = alloc(tuples * sizeof(uint32_t));
	uint64_t *bitmap = alloc(bitmasks * sizeof(uint64_t));
	for (bits = min_bits ; bits <= max_bits ; ++bits) {
		// generate min and max value
		uint32_t value_1 = rand32_next(gen) >> (32 - bits);
		uint32_t value_2 = rand32_next(gen) >> (32 - bits);
		uint32_t min_value = min(value_1, value_2);
		uint32_t max_value = max(value_1, value_2);
		// group size
		size_t group = 64 / (bits + 1);
		size_t tuples_aligned = tuples - tuples % group;
		// generate data
		for (i = 0 ; i != tuples ; ++i)
			original[i] = rand32_next(gen) >> (32 - bits);
#ifdef _NO
		compress_anyway(original, compressed, tuples_aligned, bits);
#endif
		pthread_barrier_wait(barrier++);
		// compress
		uint64_t t1 = thread_time();
		compress(original, compressed, tuples_aligned, bits);
		t1 = thread_time() - t1;
		pthread_barrier_wait(barrier++);
		// decompress
		uint64_t t2 = thread_time();
		decompress(compressed, decompressed, tuples_aligned, bits);
		t2 = thread_time() - t2;
		pthread_barrier_wait(barrier++);
		// generate bitmap (c1 <= A.x <= c2)
		uint64_t t3 = thread_time();
		compare(compressed, bitmap, tuples_aligned, bits, min_value, max_value);
		t3 = thread_time() - t3;
		pthread_barrier_wait(barrier++);
#ifndef _NO
		// verify decompression
		for (i = 0 ; i != tuples_aligned ; ++i)
			assert(original[i] == decompressed[i]);
#endif
		// count set bits
		size_t sb = 0;
		for (i = 0 ; i != (tuples_aligned + 63) >> 6 ; ++i)
			sb += _mm_popcnt_u64(bitmap[i]);
		// verify comparison
		for (i = 0 ; i != tuples_aligned ; ++i) {
			int8_t x = (original[i] >= min_value ? 1 : 0) &
			           (original[i] <= max_value ? 1 : 0);
			int8_t y = (bitmap[i >> 6] & (1ull << (i & 63))) ? 1 : 0;
//			if (x != y)
//				fprintf(stderr, "Comparison error in %ld / %ld\n", i + 1, tuples);
			assert(x == y);
		}
		d->times[0][d->thread] = t1;
		d->times[1][d->thread] = t2;
		d->times[2][d->thread] = t3;
		d->set_bits[d->thread] = sb;
		pthread_barrier_wait(barrier++);
		if (d->thread == 0) {
			t1 = t2 = t3 = sb = 0;
			for (t = 0 ; t != d->threads ; ++t) {
				t1 += d->times[0][t];
				t2 += d->times[1][t];
				t3 += d->times[2][t];
				sb += d->set_bits[t];
			}
			printf("%2d, %6.3f, %6.3f, %6.3f,, %5.2f%%\n", bits,
			       tuples * d->threads * 1.0 / (t1 / d->threads),
			       tuples * d->threads * 1.0 / (t2 / d->threads),
			       tuples * d->threads * 1.0 / (t3 / d->threads),
			       sb * 100.0 / (tuples * d->threads));
		}
	}
	free(gen);
	free(original);
	free(compressed);
	free(decompressed);
	free(bitmap);
	pthread_exit(NULL);
}

int main(int argc, char **argv)
{
	int min_bits = argc > 1 ? atoi(argv[1]) : 1;
	int max_bits = argc > 2 ? atoi(argv[2]) : 31;
	int t, threads = argc > 3 ? atoi(argv[3]) : hardware_threads();
	size_t tuples = argc > 4 ? atoll(argv[4]) : 1000 * 1000 * 1000;
	int b, barriers = (max_bits - min_bits + 1) * 7;
	pthread_barrier_t barrier[barriers];
	for (b = 0 ; b != barriers ; ++b)
		pthread_barrier_init(&barrier[b], NULL, threads);
	srand(time(NULL));
	fprintf(stderr, "Horizontal! (word aligned)\n");
	fprintf(stderr, "Threads: %d\n", threads);
#ifdef _NO
	fprintf(stderr, "No materialization!\n");
#endif
	info_t info[threads];
	uint64_t times[3][threads];
	size_t set_bits[threads];
	for (t = 0 ; t != threads ; ++t) {
		info[t].tuples = (tuples / threads) & -64;
		info[t].seed = rand();
		info[t].thread = t;
		info[t].threads = threads;
		info[t].barrier = barrier;
		info[t].min_bits = min_bits;
		info[t].max_bits = max_bits;
		info[t].set_bits = set_bits;
		info[t].times[0] = times[0];
		info[t].times[1] = times[1];
		info[t].times[2] = times[2];
		pthread_create(&info[t].id, NULL, run, (void*) &info[t]);
	}
	for (t = 0 ; t != threads ; ++t)
		pthread_join(info[t].id, NULL);
	for (b = 0 ; b != barriers ; ++b)
		pthread_barrier_destroy(&barrier[b]);
	return EXIT_SUCCESS;
}
