/* Copyright (c) 2015
 * The Trustees of Columbia University in the City of New York
 * All rights reserved.
 *
 * Author:  Orestis Polychroniou  (orestis@cs.columbia.edu)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <time.h>
#include <stdio.h>
#include <sched.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <pthread.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <immintrin.h>

#include "rand.h"


#ifdef __INTEL_COMPILER
typedef long si64;
#else
typedef long long si64;
#endif


uint64_t thread_time(void)
{
	struct timespec t;
	assert(clock_gettime(CLOCK_THREAD_CPUTIME_ID, &t) == 0);
	return t.tv_sec * 1000 * 1000 * 1000 + t.tv_nsec;
}

int hardware_threads(void)
{
	char name[64];
	struct stat st;
	int threads = -1;
	do {
		sprintf(name, "/sys/devices/system/cpu/cpu%d", ++threads);
	} while (stat(name, &st) == 0);
	return threads;
}

void bind_thread(int thread, int threads)
{
	assert(thread >= 0 && thread < threads);
	size_t size = CPU_ALLOC_SIZE(threads);
	cpu_set_t *cpu_set = CPU_ALLOC(threads);
	assert(cpu_set != NULL);
	CPU_ZERO_S(size, cpu_set);
	CPU_SET_S(thread, size, cpu_set);
	assert(pthread_setaffinity_np(pthread_self(), size, cpu_set) == 0);
	CPU_FREE(cpu_set);
}

void *alloc(uint64_t bytes)
{
	void *ptr = NULL;
	assert(posix_memalign(&ptr, 64, bytes) == 0);
	memset(ptr, 0xDB, bytes);
	return ptr;
}

uint32_t min(uint32_t x, uint32_t y) { return x < y ? x : y; }
uint32_t max(uint32_t x, uint32_t y) { return x > y ? x : y; }

void compare_scalar(const uint32_t *data, uint64_t *bitmap, size_t tuples,
                    uint32_t min, uint32_t max)
{
	const uint32_t *data_end = &data[tuples];
	uint64_t word = 0;
	if (tuples) do {
		const uint32_t *data_inner_beg = &data[63];
		const uint32_t *data_inner_end = &data[-1];
		do {
			uint32_t x = *data_inner_beg--;
			uint8_t lo = x >= min ? 1 : 0;
			uint8_t hi = x <= max ? 1 : 0;
			word += word + (uint64_t) (lo & hi);
		} while (data_inner_beg != data_inner_end);
		_mm_stream_si64((si64*) bitmap++, word);
		data += 64;
	} while (data != data_end);
}

void compare_vector(const uint32_t *data, uint64_t *bitmap, size_t tuples,
                    uint32_t min, uint32_t max)
{
	const uint32_t *data_end = &data[tuples];
	uint32_t *bitmap_word = (uint32_t*) bitmap;
	__m256i mask_order = _mm256_set_epi32(7,3,6,2,5,1,4,0);
	assert((tuples & 31) == 0);
	__m256i mask_min = _mm256_set1_epi32(min);
	__m256i mask_max = _mm256_set1_epi32(max);
	if (tuples) do {
		__m256i x1 = _mm256_load_si256((__m256i*) &data[0]);
		__m256i x2 = _mm256_load_si256((__m256i*) &data[8]);
		__m256i x3 = _mm256_load_si256((__m256i*) &data[16]);
		__m256i x4 = _mm256_load_si256((__m256i*) &data[24]);
		data += 32;
		__m256i c1_L = _mm256_cmpgt_epi32(x1, mask_max);
		__m256i c1_H = _mm256_cmpgt_epi32(mask_min, x1);
		__m256i c2_L = _mm256_cmpgt_epi32(x2, mask_max);
		__m256i c2_H = _mm256_cmpgt_epi32(mask_min, x2);
		__m256i c3_L = _mm256_cmpgt_epi32(x3, mask_max);
		__m256i c3_H = _mm256_cmpgt_epi32(mask_min, x3);
		__m256i c4_L = _mm256_cmpgt_epi32(x4, mask_max);
		__m256i c4_H = _mm256_cmpgt_epi32(mask_min, x4);
		__m256i c1 = _mm256_or_si256(c1_L, c1_H);
		__m256i c2 = _mm256_or_si256(c2_L, c2_H);
		__m256i c3 = _mm256_or_si256(c3_L, c3_H);
		__m256i c4 = _mm256_or_si256(c4_L, c4_H);
		__m256i c12 = _mm256_packs_epi32(c1, c2);
		__m256i c34 = _mm256_packs_epi32(c3, c4);
		__m256i c = _mm256_packs_epi16(c12, c34);
		c = _mm256_permutevar8x32_epi32(c, mask_order);
		uint32_t r = _mm256_movemask_epi8(c);
		_mm_stream_si32((int*) bitmap_word++, ~r);
	} while (data != data_end);
}

typedef struct {
	pthread_t id;
	int seed;
	int thread;
	int threads;
	size_t tuples;
	size_t *set_bits;
	uint64_t *times[2];
	pthread_barrier_t *barrier;
} info_t;

void *run(void *arg)
{
	info_t *d = (info_t*) arg;
	assert(pthread_equal(pthread_self(), d->id));
	bind_thread(d->thread, d->threads);
	int t;
	size_t i;
	pthread_barrier_t *barrier = d->barrier;
	size_t tuples = d->tuples;
	assert(tuples % 64 == 0);
	size_t bitmasks = tuples / 64;
	rand32_t *gen = rand32_init(d->seed);
	uint32_t *data = alloc(tuples * sizeof(uint32_t));
	uint64_t *bitmap_1 = alloc(bitmasks * sizeof(uint64_t));
	uint64_t *bitmap_2 = alloc(bitmasks * sizeof(uint64_t));
	// generate min and max value
	uint32_t value_1 = rand32_next(gen) >> 1;
	uint32_t value_2 = rand32_next(gen) >> 1;
	uint32_t min_value = min(value_1, value_2);
	uint32_t max_value = max(value_1, value_2);
	// generate data
	for (i = 0 ; i != tuples ; ++i)
		data[i] = rand32_next(gen) >> 1;
	// scalar scan
	uint64_t t1 = thread_time();
	compare_scalar(data, bitmap_1, tuples, min_value, max_value);
	t1 = thread_time() - t1;
	pthread_barrier_wait(barrier++);
	// vector scan
	uint64_t t2 = thread_time();
	compare_vector(data, bitmap_2, tuples, min_value, max_value);
	t2 = thread_time() - t2;
	pthread_barrier_wait(barrier++);
	// count set bits
	size_t sb = 0;
	for (i = 0 ; i != bitmasks ; ++i)
		sb += _mm_popcnt_u64(bitmap_1[i]);
	// verify comparison
	for (i = 0 ; i != tuples ; ++i) {
		int8_t x = (data[i] >= min_value ? 1 : 0) &
		           (data[i] <= max_value ? 1 : 0);
		int8_t y = (bitmap_1[i >> 6] & (1ull << (i & 63))) ? 1 : 0;
		int8_t z = (bitmap_2[i >> 6] & (1ull << (i & 63))) ? 1 : 0;
		assert(x == y);
		assert(x == z);
	}
	d->times[0][d->thread] = t1;
	d->times[1][d->thread] = t2;
	d->set_bits[d->thread] = sb;
	pthread_barrier_wait(barrier++);
	if (d->thread == 0) {
		t1 = t2 = sb = 0;
		for (t = 0 ; t != d->threads ; ++t) {
			t1 += d->times[0][t];
			t2 += d->times[1][t];
			sb += d->set_bits[t];
		}
		printf("%6.3f, %6.3f, %5.2f%%\n",
		       tuples * d->threads * 1.0 / (t1 / d->threads),
		       tuples * d->threads * 1.0 / (t2 / d->threads),
		       sb * 100.0 / (tuples * d->threads));
	}
	free(gen);
	free(data);
	free(bitmap_1);
	free(bitmap_2);
	pthread_exit(NULL);
}

int main(int argc, char **argv)
{
	int t, threads = argc > 1 ? atoi(argv[1]) : hardware_threads();
	size_t tuples = argc > 2 ? atoll(argv[2]) : 1000 * 1000 * 1000;
	int b, barriers = 8;
	pthread_barrier_t barrier[barriers];
	for (b = 0 ; b != barriers ; ++b)
		pthread_barrier_init(&barrier[b], NULL, threads);
	srand(time(NULL));
	fprintf(stderr, "Raw!\n");
	fprintf(stderr, "Threads: %d\n", threads);
	info_t info[threads];
	uint64_t times[2][threads];
	size_t set_bits[threads];
	for (t = 0 ; t != threads ; ++t) {
		info[t].tuples = (tuples / threads) & -64;
		info[t].seed = rand();
		info[t].thread = t;
		info[t].threads = threads;
		info[t].barrier = barrier;
		info[t].set_bits = set_bits;
		info[t].times[0] = times[0];
		info[t].times[1] = times[1];
		pthread_create(&info[t].id, NULL, run, (void*) &info[t]);
	}
	for (t = 0 ; t != threads ; ++t)
		pthread_join(info[t].id, NULL);
	for (b = 0 ; b != barriers ; ++b)
		pthread_barrier_destroy(&barrier[b]);
	return EXIT_SUCCESS;
}
