/* Copyright (c) 2015
 * The Trustees of Columbia University in the City of New York
 * All rights reserved.
 *
 * Author:  Orestis Polychroniou  (orestis@cs.columbia.edu)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <time.h>
#include <stdio.h>
#include <sched.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <pthread.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <immintrin.h>

#include "rand.h"


uint64_t thread_time(void)
{
	struct timespec t;
	assert(clock_gettime(CLOCK_THREAD_CPUTIME_ID, &t) == 0);
	return t.tv_sec * 1000 * 1000 * 1000 + t.tv_nsec;
}

int hardware_threads(void)
{
	char name[64];
	struct stat st;
	int threads = -1;
	do {
		sprintf(name, "/sys/devices/system/cpu/cpu%d", ++threads);
	} while (stat(name, &st) == 0);
	return threads;
}

void bind_thread(int thread, int threads)
{
	assert(thread >= 0 && thread < threads);
	size_t size = CPU_ALLOC_SIZE(threads);
	cpu_set_t *cpu_set = CPU_ALLOC(threads);
	assert(cpu_set != NULL);
	CPU_ZERO_S(size, cpu_set);
	CPU_SET_S(thread, size, cpu_set);
	assert(pthread_setaffinity_np(pthread_self(), size, cpu_set) == 0);
	CPU_FREE(cpu_set);
}

void *alloc(uint64_t bytes)
{
	void *ptr = NULL;
	assert(posix_memalign(&ptr, 64, bytes) == 0);
	memset(ptr, 0xDB, bytes);
	return ptr;
}

uint64_t *align(uint64_t *p)
{
	while (63 & (size_t) p) p++;
	return p;
}

uint32_t min(uint32_t x, uint32_t y) { return x < y ? x : y; }
uint32_t max(uint32_t x, uint32_t y) { return x > y ? x : y; }

uint32_t _mm256_reduce_xor_epi32(__m256i x)
{
	size_t i;
	uint32_t a[8];
	_mm256_storeu_si256((__m256i*) a, x);
	for (i = 1 ; i != 8 ; ++i)
		a[0] ^= a[i];
	return a[0];
}

uint64_t _mm256_movemask_epi8_64(__m256i x)
{
	uint64_t r;
	asm("vpmovmskb	%1, %0" : "=r"(r) : "x"(x));
	return r;
}

void compress_anyway(const uint32_t *decompressed, uint64_t *compressed, size_t tuples, int8_t bits)
{
	assert(tuples % 8192 == 0);
	const uint32_t *decompressed_end = &decompressed[tuples];
	__m128i mask_shift = _mm_cvtsi32_si128(32 - bits);
	__m256i mask_order = _mm256_set_epi32(7,3,6,2,5,1,4,0);
	size_t bits_128 = bits * (size_t) 128;
	uint64_t buffer_space[bits_128 + 7];
	uint64_t *buffer = align(buffer_space);
	if (tuples) do {
		size_t i = 0;
		do {
			size_t j = i;
			__m256i v1 = _mm256_load_si256((__m256i*) &decompressed[0]);
			__m256i v2 = _mm256_load_si256((__m256i*) &decompressed[8]);
			__m256i v3 = _mm256_load_si256((__m256i*) &decompressed[16]);
			__m256i v4 = _mm256_load_si256((__m256i*) &decompressed[24]);
			__m256i v5 = _mm256_load_si256((__m256i*) &decompressed[32]);
			__m256i v6 = _mm256_load_si256((__m256i*) &decompressed[40]);
			__m256i v7 = _mm256_load_si256((__m256i*) &decompressed[48]);
			__m256i v8 = _mm256_load_si256((__m256i*) &decompressed[56]);
			decompressed += 64;
			v1 = _mm256_sll_epi32(v1, mask_shift);
			v2 = _mm256_sll_epi32(v2, mask_shift);
			v3 = _mm256_sll_epi32(v3, mask_shift);
			v4 = _mm256_sll_epi32(v4, mask_shift);
			v5 = _mm256_sll_epi32(v5, mask_shift);
			v6 = _mm256_sll_epi32(v6, mask_shift);
			v7 = _mm256_sll_epi32(v7, mask_shift);
			v8 = _mm256_sll_epi32(v8, mask_shift);
			int8_t bit = 0;
			for (;;) {
				__m256i r1 = _mm256_srli_epi32(v1, 24);
				__m256i r2 = _mm256_srli_epi32(v2, 24);
				__m256i r3 = _mm256_srli_epi32(v3, 24);
				__m256i r4 = _mm256_srli_epi32(v4, 24);
				__m256i r12 = _mm256_packus_epi32(r1, r2);
				__m256i r34 = _mm256_packus_epi32(r3, r4);
				__m256i r5 = _mm256_srli_epi32(v5, 24);
				__m256i r6 = _mm256_srli_epi32(v6, 24);
				__m256i r7 = _mm256_srli_epi32(v7, 24);
				__m256i r8 = _mm256_srli_epi32(v8, 24);
				__m256i r56 = _mm256_packus_epi32(r5, r6);
				__m256i r78 = _mm256_packus_epi32(r7, r8);
				__m256i r1234 = _mm256_packus_epi16(r12, r34);
				__m256i r5678 = _mm256_packus_epi16(r56, r78);
				r1234 = _mm256_permutevar8x32_epi32(r1234, mask_order);
				r5678 = _mm256_permutevar8x32_epi32(r5678, mask_order);
				int8_t pass_bits = bits - bit;
				if (pass_bits > 8) pass_bits = 8;
				bit += pass_bits;
				do {
					uint64_t lo = _mm256_movemask_epi8_64(r1234);
					uint64_t hi = _mm256_movemask_epi8_64(r5678);
					r1234 = _mm256_add_epi8(r1234, r1234);
					r5678 = _mm256_add_epi8(r5678, r5678);
					buffer[j] = lo | (hi << 32);
					j += 128;
				} while (--pass_bits);
				if (bit == bits) break;
				v1 = _mm256_slli_epi32(v1, 8);
				v2 = _mm256_slli_epi32(v2, 8);
				v3 = _mm256_slli_epi32(v3, 8);
				v4 = _mm256_slli_epi32(v4, 8);
				v5 = _mm256_slli_epi32(v5, 8);
				v6 = _mm256_slli_epi32(v6, 8);
				v7 = _mm256_slli_epi32(v7, 8);
				v8 = _mm256_slli_epi32(v8, 8);
			}
		} while (++i != 128);
		uint64_t *compressed_end = &compressed[bits_128];
		uint64_t *buffer_ptr = buffer;
		do {
			__m256i x = _mm256_load_si256((__m256i*) buffer_ptr);
			_mm256_stream_si256((__m256i*) compressed, x);
			buffer_ptr += 4;
			compressed += 4;
		} while (compressed != compressed_end);
	} while (decompressed != decompressed_end);
}

void compress(const uint32_t *decompressed, uint64_t *compressed, size_t tuples, int8_t bits)
{
	assert(tuples % 8192 == 0);
	const uint32_t *decompressed_end = &decompressed[tuples];
	__m128i mask_shift = _mm_cvtsi32_si128(32 - bits);
	__m256i mask_order = _mm256_set_epi32(7,3,6,2,5,1,4,0);
#ifndef _NO
	size_t bits_128 = bits * (size_t) 128;
	uint64_t buffer_space[bits_128 + 7];
	uint64_t *buffer = align(buffer_space);
#endif
	uint64_t xor = 0;
	if (tuples) do {
		size_t i = 0;
		do {
			size_t j = i;
			__m256i v1 = _mm256_load_si256((__m256i*) &decompressed[0]);
			__m256i v2 = _mm256_load_si256((__m256i*) &decompressed[8]);
			__m256i v3 = _mm256_load_si256((__m256i*) &decompressed[16]);
			__m256i v4 = _mm256_load_si256((__m256i*) &decompressed[24]);
			__m256i v5 = _mm256_load_si256((__m256i*) &decompressed[32]);
			__m256i v6 = _mm256_load_si256((__m256i*) &decompressed[40]);
			__m256i v7 = _mm256_load_si256((__m256i*) &decompressed[48]);
			__m256i v8 = _mm256_load_si256((__m256i*) &decompressed[56]);
			decompressed += 64;
			v1 = _mm256_sll_epi32(v1, mask_shift);
			v2 = _mm256_sll_epi32(v2, mask_shift);
			v3 = _mm256_sll_epi32(v3, mask_shift);
			v4 = _mm256_sll_epi32(v4, mask_shift);
			v5 = _mm256_sll_epi32(v5, mask_shift);
			v6 = _mm256_sll_epi32(v6, mask_shift);
			v7 = _mm256_sll_epi32(v7, mask_shift);
			v8 = _mm256_sll_epi32(v8, mask_shift);
			int8_t bit = 0;
			for (;;) {
				__m256i r1 = _mm256_srli_epi32(v1, 24);
				__m256i r2 = _mm256_srli_epi32(v2, 24);
				__m256i r3 = _mm256_srli_epi32(v3, 24);
				__m256i r4 = _mm256_srli_epi32(v4, 24);
				__m256i r12 = _mm256_packus_epi32(r1, r2);
				__m256i r34 = _mm256_packus_epi32(r3, r4);
				__m256i r5 = _mm256_srli_epi32(v5, 24);
				__m256i r6 = _mm256_srli_epi32(v6, 24);
				__m256i r7 = _mm256_srli_epi32(v7, 24);
				__m256i r8 = _mm256_srli_epi32(v8, 24);
				__m256i r56 = _mm256_packus_epi32(r5, r6);
				__m256i r78 = _mm256_packus_epi32(r7, r8);
				__m256i r1234 = _mm256_packus_epi16(r12, r34);
				__m256i r5678 = _mm256_packus_epi16(r56, r78);
				r1234 = _mm256_permutevar8x32_epi32(r1234, mask_order);
				r5678 = _mm256_permutevar8x32_epi32(r5678, mask_order);
				int8_t pass_bits = bits - bit;
				if (pass_bits > 8) pass_bits = 8;
				bit += pass_bits;
				do {
					uint64_t lo = _mm256_movemask_epi8_64(r1234);
					uint64_t hi = _mm256_movemask_epi8_64(r5678);
					r1234 = _mm256_add_epi8(r1234, r1234);
					r5678 = _mm256_add_epi8(r5678, r5678);
					uint64_t lh = lo | (hi << 32);
#ifdef _NO
					xor |= lh;
#else
					buffer[j] = lh;
#endif
					j += 128;
				} while (--pass_bits);
				if (bit == bits) break;
				v1 = _mm256_slli_epi32(v1, 8);
				v2 = _mm256_slli_epi32(v2, 8);
				v3 = _mm256_slli_epi32(v3, 8);
				v4 = _mm256_slli_epi32(v4, 8);
				v5 = _mm256_slli_epi32(v5, 8);
				v6 = _mm256_slli_epi32(v6, 8);
				v7 = _mm256_slli_epi32(v7, 8);
				v8 = _mm256_slli_epi32(v8, 8);
			}
		} while (++i != 128);
#ifndef _NO
		uint64_t *compressed_end = &compressed[bits_128];
		uint64_t *buffer_ptr = buffer;
		do {
			__m256i x = _mm256_load_si256((__m256i*) buffer_ptr);
			_mm256_stream_si256((__m256i*) compressed, x);
			buffer_ptr += 4;
			compressed += 4;
		} while (compressed != compressed_end);
#endif
	} while (decompressed != decompressed_end);
	if (xor == 666)
		fprintf(stderr, "666 the number of the beast!\n");
}

void decompress(const uint64_t *compressed, uint32_t *decompressed, size_t tuples, int8_t bits)
{
	assert(tuples % 8192 == 0);
	const uint64_t *compressed_end = &compressed[(tuples >> 6) * bits];
	const uint64_t mask_4 = 0x0404040404040404ull;
	const uint64_t mask_8 = 0x0808080808080808ull;
	const uint64_t mask_C = 0x0C0C0C0C0C0C0C0Cull;
	const uint64_t mask_P = 0x8040201008040201ull;
	__m256i mask_offset = _mm256_set1_epi64x(mask_P);
	__m256i mask_broadcast_lo = _mm256_set_epi64x(mask_8,      0, mask_8,      0);
	__m256i mask_broadcast_hi = _mm256_set_epi64x(mask_C, mask_4, mask_C, mask_4);
	__m256i xor = _mm256_setzero_si256();
	size_t bits_128 = bits * (size_t) 128;
	if (tuples) do {
		size_t i = 0;
		do {
			__m256i v1 = _mm256_setzero_si256();
			__m256i v2 = _mm256_setzero_si256();
			__m256i v3 = _mm256_setzero_si256();
			__m256i v4 = _mm256_setzero_si256();
			__m256i v5 = _mm256_setzero_si256();
			__m256i v6 = _mm256_setzero_si256();
			__m256i v7 = _mm256_setzero_si256();
			__m256i v8 = _mm256_setzero_si256();
			int8_t bit = bits;
			size_t j = i;
			for (;;) {
				__m256i x = _mm256_setzero_si256();
				__m256i y = _mm256_setzero_si256();
				do {
					__m128i lhc = _mm_loadl_epi64((__m128i*) &compressed[j]);
					__m256i lh = _mm256_cvtepu8_epi32(lhc);
					j += 128;
					__m256i lo = _mm256_shuffle_epi8(lh, mask_broadcast_lo);
					__m256i hi = _mm256_shuffle_epi8(lh, mask_broadcast_hi);
					lo = _mm256_and_si256(lo, mask_offset);
					hi = _mm256_and_si256(hi, mask_offset);
					x = _mm256_add_epi8(x, x);
					y = _mm256_add_epi8(y, y);
					lo = _mm256_cmpeq_epi8(lo, mask_offset);
					hi = _mm256_cmpeq_epi8(hi, mask_offset);
					x = _mm256_sub_epi8(x, lo);
					y = _mm256_sub_epi8(y, hi);
				} while (--bit & 7);
				__m128i a = _mm256_castsi256_si128(x);
				__m128i b = _mm256_castsi256_si128(y);
				v1 = _mm256_or_si256(v1, _mm256_cvtepu8_epi32(a));
				v2 = _mm256_or_si256(v2, _mm256_cvtepu8_epi32(b));
				a = _mm_srli_si128(a, 8);
				b = _mm_srli_si128(b, 8);
				v3 = _mm256_or_si256(v3, _mm256_cvtepu8_epi32(a));
				v4 = _mm256_or_si256(v4, _mm256_cvtepu8_epi32(b));
				x = _mm256_permute4x64_epi64(x, _MM_SHUFFLE(3,2,3,2));
				y = _mm256_permute4x64_epi64(y, _MM_SHUFFLE(3,2,3,2));
				__m128i c = _mm256_castsi256_si128(x);
				__m128i d = _mm256_castsi256_si128(y);
				v5 = _mm256_or_si256(v5, _mm256_cvtepu8_epi32(c));
				v6 = _mm256_or_si256(v6, _mm256_cvtepu8_epi32(d));
				c = _mm_srli_si128(c, 8);
				d = _mm_srli_si128(d, 8);
				v7 = _mm256_or_si256(v7, _mm256_cvtepu8_epi32(c));
				v8 = _mm256_or_si256(v8, _mm256_cvtepu8_epi32(d));
				if (bit == 0) break;
				v1 = _mm256_slli_epi32(v1, 8);
				v2 = _mm256_slli_epi32(v2, 8);
				v3 = _mm256_slli_epi32(v3, 8);
				v4 = _mm256_slli_epi32(v4, 8);
				v5 = _mm256_slli_epi32(v5, 8);
				v6 = _mm256_slli_epi32(v6, 8);
				v7 = _mm256_slli_epi32(v7, 8);
				v8 = _mm256_slli_epi32(v8, 8);
			}
#ifdef _NO
			v1 = _mm256_xor_si256(v1, v5);
			v2 = _mm256_xor_si256(v2, v6);
			v3 = _mm256_xor_si256(v3, v7);
			v4 = _mm256_xor_si256(v4, v8);
			v1 = _mm256_xor_si256(v1, v3);
			v2 = _mm256_xor_si256(v2, v4);
			v1 = _mm256_xor_si256(v1, v2);
			xor = _mm256_xor_si256(xor, v1);
#else
			_mm256_stream_si256((__m256i*) &decompressed[ 0], v1);
			_mm256_stream_si256((__m256i*) &decompressed[ 8], v2);
			_mm256_stream_si256((__m256i*) &decompressed[16], v3);
			_mm256_stream_si256((__m256i*) &decompressed[24], v4);
			_mm256_stream_si256((__m256i*) &decompressed[32], v5);
			_mm256_stream_si256((__m256i*) &decompressed[40], v6);
			_mm256_stream_si256((__m256i*) &decompressed[48], v7);
			_mm256_stream_si256((__m256i*) &decompressed[56], v8);
			decompressed += 64;
#endif
		} while (++i != 128);
		compressed += bits_128;
	} while (compressed != compressed_end);
	if (_mm256_reduce_xor_epi32(xor) == 666)
		fprintf(stderr, "666 the number of the beast!\n");
}

void compare(const uint64_t *compressed, uint64_t *bitmap, size_t tuples,
             int8_t bits, uint32_t min, uint32_t max)
{
	assert(tuples % 8192 == 0);
	const uint64_t *compressed_end = &compressed[(tuples * bits) >> 6];
	__m256i min_shift = _mm256_set1_epi32(min << (32 - bits));
	__m256i max_shift = _mm256_set1_epi32(max << (32 - bits));
	__m256i all_set = _mm256_set1_epi32(-1);
	size_t bits_128 = bits * (size_t) 128;
	if (tuples) do {
		size_t i = 0;
		do {
			__m256i min_tmp = min_shift;
			__m256i max_tmp = max_shift;
			__m256i gt_min = _mm256_setzero_si256();
			__m256i lt_max = _mm256_setzero_si256();
			__m256i eq_min = all_set;
			__m256i eq_max = all_set;
			int8_t b = bits;
			size_t j = i;
			do {
				__m256i data = _mm256_load_si256((__m256i*) &compressed[j]);
				__m256i min_bit = _mm256_srai_epi32(min_tmp, 31);
				__m256i max_bit = _mm256_srai_epi32(max_tmp, 31);
				__m256i xor_min = _mm256_xor_si256(min_bit, data);
				__m256i xor_max = _mm256_xor_si256(max_bit, data);
				__m256i data_xor_min = _mm256_and_si256(data, xor_min);
				__m256i data_xor_max = _mm256_andnot_si256(data, xor_max);
				data_xor_min = _mm256_and_si256(data_xor_min, eq_min);
				data_xor_max = _mm256_and_si256(data_xor_max, eq_max);
				gt_min = _mm256_or_si256(gt_min, data_xor_min);
				lt_max = _mm256_or_si256(lt_max, data_xor_max);
				eq_min = _mm256_andnot_si256(xor_min, eq_min);
				eq_max = _mm256_andnot_si256(xor_max, eq_max);
				__m256i eq = _mm256_or_si256(eq_min, eq_max);
				if (_mm256_testz_si256(eq, eq)) break;
				min_tmp = _mm256_add_epi32(min_tmp, min_tmp);
				max_tmp = _mm256_add_epi32(max_tmp, max_tmp);
				j += 128;
			} while (--b);
			__m256i ge_min = _mm256_or_si256(gt_min, eq_min);
			__m256i le_max = _mm256_or_si256(lt_max, eq_max);
			__m256i r = _mm256_and_si256(ge_min, le_max);
			_mm256_stream_si256((__m256i*) &bitmap[i], r);
			i += 4;
		} while (i != 128);
		compressed += bits_128;
		bitmap += 128;
	} while (compressed != compressed_end);
}

typedef struct {
	pthread_t id;
	int seed;
	int thread;
	int threads;
	int min_bits;
	int max_bits;
	size_t tuples;
	size_t *set_bits;
	uint64_t *times[3];
	pthread_barrier_t *barrier;
} info_t;

void *run(void *arg)
{
	info_t *d = (info_t*) arg;
	assert(pthread_equal(pthread_self(), d->id));
	bind_thread(d->thread, d->threads);
	int bits, t;
	int min_bits = d->min_bits;
	int max_bits = d->max_bits;
	size_t i;
	pthread_barrier_t *barrier = d->barrier;
	size_t tuples = d->tuples;
	assert(tuples % 8192 == 0);
	size_t bitmasks = tuples / 64;
	size_t words = bitmasks * max_bits;
	rand32_t *gen = rand32_init(d->seed);
	uint32_t *original = alloc(tuples * sizeof(uint32_t));
	uint64_t *compressed = alloc(words * sizeof(uint64_t));
	uint32_t *decompressed = alloc(tuples * sizeof(uint32_t));
	uint64_t *bitmap = alloc(bitmasks * sizeof(uint64_t));
	for (bits = min_bits ; bits <= max_bits ; ++bits) {
		// generate min and max value
		uint32_t value_1 = rand32_next(gen) >> (32 - bits);
		uint32_t value_2 = rand32_next(gen) >> (32 - bits);
		uint32_t min_value = min(value_1, value_2);
		uint32_t max_value = max(value_1, value_2);
		// generate data
		for (i = 0 ; i != tuples ; ++i)
			original[i] = rand32_next(gen) >> (32 - bits);
#ifdef _NO
		compress_anyway(original, compressed, tuples, bits);
#endif
		pthread_barrier_wait(barrier++);
		// compress
		uint64_t t1 = thread_time();
		compress(original, compressed, tuples, bits);
		t1 = thread_time() - t1;
		pthread_barrier_wait(barrier++);
		// decompress
		uint64_t t2 = thread_time();
		decompress(compressed, decompressed, tuples, bits);
		t2 = thread_time() - t2;
		pthread_barrier_wait(barrier++);
		// generate bitmap (c1 <= A.x <= c2)
		uint64_t t3 = thread_time();
		compare(compressed, bitmap, tuples, bits, min_value, max_value);
		t3 = thread_time() - t3;
		pthread_barrier_wait(barrier++);
#ifndef _NO
		// verify decompression
		for (i = 0 ; i != tuples ; ++i)
			assert(original[i] == decompressed[i]);
#endif
		// count set bits
		size_t sb = 0;
		for (i = 0 ; i != bitmasks ; ++i)
			sb += _mm_popcnt_u64(bitmap[i]);
		// verify comparison
		for (i = 0 ; i != tuples ; ++i) {
			int8_t x = (original[i] >= min_value ? 1 : 0) &
			           (original[i] <= max_value ? 1 : 0);
			int8_t y = (bitmap[i >> 6] & (1ull << (i & 63))) ? 1 : 0;
			assert(x == y);
		}
		d->times[0][d->thread] = t1;
		d->times[1][d->thread] = t2;
		d->times[2][d->thread] = t3;
		d->set_bits[d->thread] = sb;
		pthread_barrier_wait(barrier++);
		if (d->thread == 0) {
			t1 = t2 = t3 = sb = 0;
			for (t = 0 ; t != d->threads ; ++t) {
				t1 += d->times[0][t];
				t2 += d->times[1][t];
				t3 += d->times[2][t];
				sb += d->set_bits[t];
			}
			printf("%2d, %6.3f, %6.3f, %6.3f,, %5.2f%%\n", bits,
			       tuples * d->threads * 1.0 / (t1 / d->threads),
			       tuples * d->threads * 1.0 / (t2 / d->threads),
			       tuples * d->threads * 1.0 / (t3 / d->threads),
			       sb * 100.0 / (tuples * d->threads));
		}
	}
	free(gen);
	free(original);
	free(compressed);
	free(decompressed);
	free(bitmap);
	pthread_exit(NULL);
}

int main(int argc, char **argv)
{
	int min_bits = argc > 1 ? atoi(argv[1]) : 1;
	int max_bits = argc > 2 ? atoi(argv[2]) : 32;
	int t, threads = argc > 3 ? atoi(argv[3]) : hardware_threads();
	size_t tuples = argc > 4 ? atoll(argv[4]) : 1000013824;
	int b, barriers = (max_bits - min_bits + 1) * 7;
	pthread_barrier_t barrier[barriers];
	for (b = 0 ; b != barriers ; ++b)
		pthread_barrier_init(&barrier[b], NULL, threads);
	srand(time(NULL));
	fprintf(stderr, "Vertical!\n");
	fprintf(stderr, "Threads: %d\n", threads);
	fprintf(stderr, "SIMD instructions! (AVX2)\n");
#ifdef _NO
	fprintf(stderr, "No materialization!\n");
#endif
	info_t info[threads];
	uint64_t times[3][threads];
	size_t set_bits[threads];
	for (t = 0 ; t != threads ; ++t) {
		info[t].tuples = (tuples / threads) & -8192;
		info[t].seed = rand();
		info[t].thread = t;
		info[t].threads = threads;
		info[t].barrier = barrier;
		info[t].min_bits = min_bits;
		info[t].max_bits = max_bits;
		info[t].set_bits = set_bits;
		info[t].times[0] = times[0];
		info[t].times[1] = times[1];
		info[t].times[2] = times[2];
		pthread_create(&info[t].id, NULL, run, (void*) &info[t]);
	}
	for (t = 0 ; t != threads ; ++t)
		pthread_join(info[t].id, NULL);
	for (b = 0 ; b != barriers ; ++b)
		pthread_barrier_destroy(&barrier[b]);
	return EXIT_SUCCESS;
}
